//
// file :         readthread.cpp
//
// description :  reading thread
//
// project :      TANGO Device Server
//
// $Author: graziano $ 
//
// $Revision: 1.13 $
//
//
//
// copyleft :   Sincrotrone Trieste S.C.p.A.
//              Strada Statale 14 - km 163,5 in AREA Science Park
//              34012 Basovizza, Trieste ITALY
//

#include "readthread.h"
#ifdef HAVE_RNM
#include "k_rnmshare.h"
#endif

static const char __FILE__rev[] = __FILE__ " $Revision: 1.13 $";

//+------------------------------------------------------------------
//
//	method:			readthread::readthread()
//
//	description:	readthread constructor
//
//-------------------------------------------------------------------
readthread::readthread(OceanOptics_ns::OceanOptics *s):Tango::LogAdapter(s)
{ 
	DEBUG_STREAM << "readthread::readthread(): constructor... !" << endl;
	mysolar = s;
	//mutex_sleep = new omni_mutex();
	min_fit_time = 1000000000.0;
	max_fit_time = 0;
	avg_fit_time = 0;
	tot_fit_time = 0;
	fit_counter = 0;
	use_last_fit = false;
#ifdef HAVE_RNM
	int ret = rnmshare_init(0);
#endif
}

//+------------------------------------------------------------------
//
//	method:			readthread::~readthread()
//
//	description:	readthread destructor
//
//-------------------------------------------------------------------
readthread::~readthread()
{
	DEBUG_STREAM << "readthread::~readthread(): destructor... !" << endl;
	//delete mutex_sleep;
#ifdef HAVE_RNM
	rnmshare_deinit();
#endif
}

//+------------------------------------------------------------------
//
//	method:			readthread::run()
//
//	description:	Run
//
//-------------------------------------------------------------------
void readthread::run(void *) 
{
	INFO_STREAM << "readthread::run(): running... !" << endl;	
	int cnt_err = 0,
		max_cnt_err = 3;
//	im540_map::iterator it;// = myim540->read_data.begin();
	
	string id;
	//update_scheduler();
	string resp;
	//mutex_sleep->lock();
	pausesec = 1;
	pausenano = 0;
	//mutex_sleep->unlock();
	int counter = 0;
	bool ret;
	double *pNoiseData=NULL;
	int flag;
	int error;
	char error_str[128];
	just_reconnected = false;
	
	while (!mysolar->abortflag) 
	{
		if(mysolar->device_id == -1)
		{
			mysolar->connect_device(true);

			if(mysolar->device_id == -1)
			{
				abort_sleep(pausesec + ((double)pausenano)/1000000000);
				continue;
			}
			else
			{
				just_reconnected = true;
			}
		}
		//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
		//+++++++++++++++++++++++++++++INIT MODE+++++++++++++++++++++++++++++++//
		//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
		if(mysolar->init_live_mode || mysolar->init_measure_mode || mysolar->init_acquire_noise_mode)
		{
			wait_counter = 0;
			/*ret = mysolar->m_usbCCD.GetParameters();
			if(!ret)
			{
				INFO_STREAM << "readthread::run(): error in GetParameters()" << endl;
				mysolar->init_live_mode = false;
				mysolar->init_measure_mode = false;
				mysolar->init_acquire_noise_mode = false;
				mysolar->live_mode = false;
				mysolar->measure_mode = false;
				mysolar->acquire_noise_mode = false;
				mysolar->set_state(Tango::ON);
				mysolar->set_status("Error Initializing Acquisition");
				continue;
			}
			else*/
			{
				//mysolar->NumPixel = mysolar->m_usbCCD.m_nNumPixelsH * mysolar->m_usbCCD.m_nNumPixelsV;
#if _SEABREEZE_API > 1
				mysolar->NumPixel = sbapi_spectrometer_get_formatted_spectrum_length(mysolar->device_id,mysolar->spectrometer_id, &error);
#else
				mysolar->NumPixel = seabreeze_get_formatted_spectrum_length(mysolar->device_id, &error);
#endif
				DEBUG_STREAM << "readthread::run() allocating for Data size=" << ((int)mysolar->NumPixel)/**((int)mysolar->m_usbCCD.m_nNumReadOuts)*/ << endl;
				if((!mysolar->init_acquire_noise_mode) && mysolar->pLiveData)
				{
					delete [] mysolar->pLiveData;
					mysolar->pLiveData = NULL;
				}
				if(!mysolar->init_acquire_noise_mode)
				{
					mysolar->pLiveData = new double[((int)mysolar->NumPixel)/**((int)mysolar->m_usbCCD.m_nNumReadOuts)*/];
				}
				else
				{
					pNoiseData = new double[((int)mysolar->NumPixel)/**((int)mysolar->m_usbCCD.m_nNumReadOuts)*/ * mysolar->noise_frames];

				}
				//DEBUG_STREAM << "readthread::run() allocated for Data size=" << ((int)mysolar->NumPixel)*((int)mysolar->m_usbCCD.m_nNumReadOuts) << endl;
				/*ret= mysolar->m_usbCCD.StartMeasuring(mysolar->pLiveData);
				if(!ret)
				{
					INFO_STREAM << "readthread::run(): error in StartMeasuring()" << endl;
					mysolar->init_live_mode = false;
					mysolar->live_mode = false;
					mysolar->set_state(Tango::ON);
					mysolar->set_status("Error Starting Continuous Mode");
					continue;
				}*/
				/*if(mysolar->nFramesBuffered > 0)
				{
					mysolar->pFrameBuffer = new DWORD[mysolar->NumPixel * mysolar->nFramesBuffered];
					memset(mysolar->pFrameBuffer,0,mysolar->NumPixel * mysolar->nFramesBuffered * sizeof(DWORD));
				}*/
				//mutex_sleep->lock();
				pausesec = 0;
				//pausenano = 1000000;		//TODO: 1 msec, it should depend on freq of CCD
				//pausenano = (*(mysolar->attr_ExposureTime_read) * *(mysolar->attr_NumReadouts_read) * 300000);		//TODO: 10 msec, it should depend on freq of CCD
				pausenano = 0;
				DEBUG_STREAM << "readthread::run(): setting sleep_time(ns)="<<pausenano<<endl;
				//mutex_sleep->unlock();
				mysolar->init_live_mode = false;
				mysolar->init_measure_mode = false;
				mysolar->init_acquire_noise_mode = false;
			}
		}
		//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
		//+++++++++++++++++++++++++++ACQUIRE MODE++++++++++++++++++++++++++++++//
		//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
		if(mysolar->live_mode || mysolar->measure_mode)
		{
			DWORD stat;
			//DEBUG_STREAM << "readthread::run(): before GetMeasStatus"<<endl;
			//mysolar->m_usbCCD.GetMeasStatus(&stat);
			//if(stat == STATUS_DATA_READY)
			if(1)
			{
				wait_counter = 0;
				//if(mysolar->m_usbCCD.m_nNumReadOuts == 1)
				if(1)
				{
					if(*(mysolar->attr_NoiseSubtraction_read))
					{
						//DEBUG_STREAM << "readthread::run(): before GetData readouts=1, noise subtraction"<<endl;
						//mysolar->m_usbCCD.GetData(mysolar->pLiveData);
#if _SEABREEZE_API > 1
						/*TODO: length*/flag = sbapi_spectrometer_get_formatted_spectrum(mysolar->device_id, mysolar->spectrometer_id, &error, mysolar->pLiveData, (int)mysolar->NumPixel);
#else
						flag = seabreeze_get_formatted_spectrum(mysolar->device_id, &error, mysolar->pLiveData, (int)mysolar->NumPixel);
#endif
						DEBUG_STREAM << "readthread::run(): seabreeze_get_formatted_spectrum: -> "<<flag<<endl;
						if(flag == 0)
						{
							seabreeze_get_error_string(error, error_str, 128);
							DEBUG_STREAM << "readthread::run(): error in seabreeze_get_formatted_spectrum: '"<<error_str<<"'"<<endl;
							mysolar->set_state(Tango::FAULT);
							mysolar->set_status("Communication error");
							mysolar->disconnect_device();
							mysolar->device_id = -1;
							if(mysolar->live_mode)
								mysolar->init_live_mode = true;
							if(mysolar->measure_mode)
								mysolar->init_measure_mode = true;
						}
						for(unsigned int j=0; j < mysolar->NumPixel; j++)
						{
							//mysolar->pLiveData[j] = (((int)mysolar->pLiveData[j] - mysolar->noise_level) > 0) ? ((int)mysolar->pLiveData[j] - mysolar->noise_level) : 0;
							if(*mysolar->attr_IntensityCalibrationEnable_read)
							{
								mysolar->pLiveData_avg[j] = (mysolar->pLiveData[j] - mysolar->pNoiseMask_avg[j]);
								if(mysolar->pLiveData_avg[j] > 0)
									mysolar->pLiveData_avg[j] *= mysolar->attr_IntensityCalibration_read[j];
							}
							else
							{
								mysolar->pLiveData_avg[j] = (mysolar->pLiveData[j] - mysolar->pNoiseMask_avg[j]);
							}
							/*//TTTTTTTTTTTTTTTTEEEEEEEEEEEEEEEEESSSSSSSSSSSSSSTTTTTTTTTTTTTTTTTTTTTTTTTTTT: generate fake gaussian
							double base = 0 + (-1 + 2*(counter%2)) * counter%101;
							double amp =15000 + (1 - 2*(counter%2)) * counter%547;
							double center = 350 + (-1 + 2*(counter%2)) * counter%3;
							double width = 10  + (1 - 2*(counter%2)) * counter%2;
							mysolar->pLiveData_avg[j] += base + amp*exp((-0.5) * pow((mysolar->x_values[j]-center)/width,2));*/
						}
						//memcpy(mysolar->pLiveData_avg,mysolar->pLiveData,mysolar->NumPixel * sizeof(DWORD));
					}
					else
					{
						//DEBUG_STREAM << "readthread::run(): before GetData readouts=1"<<endl;
						//mysolar->m_usbCCD.GetData(mysolar->pLiveData);
#if _SEABREEZE_API > 1
						/*TODO: length*/flag = sbapi_spectrometer_get_formatted_spectrum(mysolar->device_id, mysolar->spectrometer_id, &error, mysolar->pLiveData, (int)mysolar->NumPixel);
#else
						flag = seabreeze_get_formatted_spectrum(mysolar->device_id, &error, mysolar->pLiveData, (int)mysolar->NumPixel);
#endif
						//mysolar->m_usbCCD.GetData(mysolar->pLiveData_avg);
						if(*mysolar->attr_IntensityCalibrationEnable_read)
						{
							for(unsigned int j=0; j < mysolar->NumPixel; j++)
								mysolar->pLiveData_avg[j] = (mysolar->attr_IntensityCalibration_read[j] * mysolar->pLiveData[j]);
						}
						else
						{
							memcpy(mysolar->pLiveData_avg,mysolar->pLiveData,mysolar->NumPixel * sizeof(double));
						}
					}
					/*for(unsigned int i=0; i<mysolar->NumPixel; i++)
					{
						DEBUG_STREAM << "OceanOptics::OceanOptics(): mysolar->pLiveData["<<i<<"]=" << mysolar->pLiveData[i] << endl;
					}*/
					DEBUG_STREAM << __func__<<": READ DATA" << endl;
					clock_gettime(CLOCK_MONOTONIC,&mysolar->last_read);
				}
				/*else
				{
					//DEBUG_STREAM << "readthread::run(): before GetData readouts>1"<<endl;
					mysolar->m_usbCCD.GetData(mysolar->pLiveData);
					//memset(&(mysolar->pLiveData_avg[0]), 0,mysolar->NumPixel*sizeof(DWORD));
				
					//for every pixel
					for(unsigned int j=0; j < mysolar->NumPixel; j++)
					{
						mysolar->pLiveData_avg[j] = 0;
						//sum value in this pixel point for every frame
						for(int i=0; i < mysolar->m_usbCCD.m_nNumReadOuts; i++)
						{
							mysolar->pLiveData_avg[j] += mysolar->pLiveData[i * mysolar->NumPixel + j];
						}
						//divide the sum per the number of frames
						mysolar->pLiveData_avg[j] = (DWORD)((((double)mysolar->pLiveData_avg[j])/((double)mysolar->m_usbCCD.m_nNumReadOuts)) + 0.5); //+0.5 to round and not truncate
						if(*(mysolar->attr_NoiseSubtraction_read))
						{
							//mysolar->pLiveData_avg[j] = (((int)mysolar->pLiveData_avg[j] - mysolar->noise_level) > 0) ? ((int)mysolar->pLiveData_avg[j] - mysolar->noise_level) : 0;
							mysolar->pLiveData_avg[j] = (int)mysolar->pLiveData_avg[j] - mysolar->noise_level;
						}
					}
				}*/
				if(mysolar->nFramesBuffered > 0)
				{
					//copy the average frame
					memcpy(&(mysolar->pFrameBuffer[mysolar->NumPixel * mysolar->FrameBufferIndex]), &(mysolar->pLiveData_avg[0]),mysolar->NumPixel*sizeof(DWORD));
					////update index in buffer once for all readout frames
					mysolar->FrameBufferIndex = (mysolar->FrameBufferIndex + 1) % mysolar->nFramesBuffered;
				}

				if(just_reconnected)
				{
					Tango::MultiAttribute *attr_list = mysolar->get_device_attr();
					try
					{
						Tango::WAttribute &attr_reg4 = attr_list->get_w_attr_by_name ("SyncMode");
						//mysolar->write_SyncMode(attr_reg4);
						Tango::DevLong w_val;
						attr_reg4.get_write_value(w_val);
						DEBUG_STREAM << "seabreeze_set_trigger_mode val=" << w_val << endl;
						seabreeze_set_trigger_mode(mysolar->device_id, &error, w_val);
						DEBUG_STREAM << "seabreeze_set_trigger_mode written , error=" << error << endl;
					}catch (...){}
					if(error == 0)
						just_reconnected = false;
					usleep(100000);
				}

				struct timeval buf_time;
				gettimeofday(&buf_time,NULL);
#if __GNUC__
#if __x86_64__ || __ppc64__
				int32_t buf_bunch_num;
#else
				long buf_bunch_num;
#endif
#else
				long buf_bunch_num;
#endif

				char check;


				if(*mysolar->attr_RtMode_read)
				{
#ifdef HAVE_RNM
					rnm_bunch_number_get_value(0, 1, &buf_bunch_num, &check);
					//buf_bunch_num = fit_counter;
#endif
					if(old_bunch_number != buf_bunch_num-1)
					{
						//INFO_STREAM << "readthread::run(): BUNCH error: previous="<<old_bunch_number<<" present="<<bunch_number;
						//cout << "--------------------------------------->>>>>>>>>>>>>>>>>>>>>>>>> BUNCH error: previous="<<old_bunch_number<<" present="<<bunch_number<<"<<<<<<<<<<<<<<<<<<<<<<<<<---------------------------"<<endl;
						INFO_STREAM << "--->>> BUNCH error: previous="<<old_bunch_number<<" present="<<buf_bunch_num<<"<<<---"<<endl;
						old_bunch_number = buf_bunch_num;
					}
					else
					{
						old_bunch_number++;
					}
				}
				else
				{
					buf_bunch_num = fit_counter;
				}
				//cout << __func__<<": before push_data data[0]="<<mysolar->pLiveData_avg[0]<<" data[1]="<<mysolar->pLiveData_avg[1]<<" data[2]="<<mysolar->pLiveData_avg[2]<<" data[1000]="<<mysolar->pLiveData_avg[1000]<<" data[1001]="<<mysolar->pLiveData_avg[1001]<<" data[1002]="<<mysolar->pLiveData_avg[1002]<<endl;
				tbuffer::_data_array<double> val_array(mysolar->pLiveData_avg, mysolar->NumPixel);
				mysolar->tbuf_array.push_data(&val_array,buf_time,buf_bunch_num);
				tbuffer::_data_array<double> val_array_w(&(mysolar->pLiveData_avg[0 + mysolar->RegionOfInterestStartPix]), mysolar->RegionOfInterestEndPix - mysolar->RegionOfInterestStartPix + 1);
				mysolar->tbuf_array_w.push_data(&val_array_w,buf_time,buf_bunch_num);
				tbuffer::_data_array<double> val_array_l(&(mysolar->x_values[0 + mysolar->RegionOfInterestStartPix]), mysolar->RegionOfInterestEndPix - mysolar->RegionOfInterestStartPix + 1);
				mysolar->tbuf_array_l.push_data(&val_array_l,buf_time,buf_bunch_num);
				//^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
				//:::::::::::::::::::::::  FWHM processing :::::::::::::::::::::::
				//________________________________________________________________
				if(*(mysolar->attr_FWHMEnable_read))
				{
					size_t x_peak, x_half_min, x_half_max;
					double y_peak;
					mysolar->fwhm_calculate(mysolar->pLiveData_avg,mysolar->NumPixel,
						mysolar->AnalysisRegionStartPix,mysolar->AnalysisRegionEndPix,
						1,				//average on 1 frame
						*(mysolar->attr_FWHMAvg_read),				//average between 3 adjacent values
						x_peak, y_peak, 
						x_half_min, x_half_max);
					mysolar->attr_FWHM_read[0]=mysolar->x_values[x_peak];
					mysolar->attr_FWHM_read[1]=y_peak;
					if(*(mysolar->attr_WidhtCorrectionEnable_read))
					{
						mysolar->attr_FWHM_read[2]=sqrt((mysolar->x_values[x_half_max] - mysolar->x_values[x_half_min])*(mysolar->x_values[x_half_max] - mysolar->x_values[x_half_min]) + mysolar->widthCorrection*mysolar->widthCorrection);
					}
					else
					{
						mysolar->attr_FWHM_read[2]=mysolar->x_values[x_half_max] - mysolar->x_values[x_half_min];
					}
					*(mysolar->attr_FWHMCenter_read) = mysolar->attr_FWHM_read[0];
					*(mysolar->attr_FWHMWidth_read) = mysolar->attr_FWHM_read[2];
					double val_buff;
					val_buff = *(mysolar->attr_FWHMCenter_read);
					mysolar->tbuf[BUF_ID_FWHM_CENTER].push_data(&val_buff,buf_time,buf_bunch_num);
					val_buff = *(mysolar->attr_FWHMWidth_read);
					mysolar->tbuf[BUF_ID_FWHM_WIDTH].push_data(&val_buff,buf_time,buf_bunch_num);
					if(!(*(mysolar->attr_GaussFitEnable_read)) || mysolar->preferFWHM)
						*(mysolar->attr_Center_read) = mysolar->attr_FWHM_read[0];
						*(mysolar->attr_Width_read) = mysolar->attr_FWHM_read[2];
				}
				//^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
				//::::::::::::::::::::::::  AUTO EXPOSURE ::::::::::::::::::::::::
				//________________________________________________________________
				if(*(mysolar->attr_AutoExposure_read))
				{
					int y_max;
					if(*(mysolar->attr_FWHMEnable_read))
					{
						y_max=(int)mysolar->attr_FWHM_read[1];
					}
					else
					{
						size_t x_peak, x_half_min, x_half_max;
						double y_peak;
						mysolar->fwhm_calculate(mysolar->pLiveData_avg,mysolar->NumPixel,
							mysolar->AnalysisRegionStartPix,mysolar->AnalysisRegionEndPix,
							1,				//average on 1 frame
							*(mysolar->attr_FWHMAvg_read),				//average between 3 adjacent values
							x_peak, y_peak, 
							x_half_min, x_half_max);
						y_max = y_peak;
					}
					mysolar->exposures[mysolar->exposure_idx] = y_max;
					mysolar->exposure_idx = (mysolar->exposure_idx + 1)%AUTO_EXPOSURE_MEAN;
					mysolar->y_max_mean = 0;
					for(int i=0; i<AUTO_EXPOSURE_MEAN; i++)
						mysolar->y_max_mean += mysolar->exposures[i];
					mysolar->y_max_mean = (int)(((double)mysolar->y_max_mean)/AUTO_EXPOSURE_MEAN);

					DEBUG_STREAM << "readthread::run(): AutoExposure: max="<<mysolar->y_max_mean<<endl;

					if(mysolar->y_max_mean > mysolar->autoExposureAmpRange[1])
					{
						double exposure = *(mysolar->attr_ExposureTime_read);
						if(exposure - mysolar->autoExposureStep >= mysolar->autoExposureRange[0])
						{
							exposure -= mysolar->autoExposureStep;
						}
						else
						{
							exposure = mysolar->autoExposureRange[0];
						}
						if(exposure != mysolar->attr_ExposureTime_write)
						{
							DEBUG_STREAM << "readthread::run(): AutoExposure: decresing to "<<exposure<<endl;
							//mysolar->m_usbCCD.SetExpTime((float)exposure);
#if _SEABREEZE_API > 1
							unsigned long integr_time_usec = (unsigned long)(exposure*1000);
							sbapi_spectrometer_set_integration_time_micros (mysolar->device_id, mysolar->spectrometer_id, &error, integr_time_usec);
#else
							long integr_time_usec = (long)(exposure*1000);
						    seabreeze_set_integration_time_microsec(mysolar->device_id, &error, integr_time_usec);
#endif
							mysolar->attr_ExposureTime_write = exposure;
						}
					}
					else if(mysolar->y_max_mean < mysolar->autoExposureAmpRange[0])
					{
						double exposure = *(mysolar->attr_ExposureTime_read);
						if(exposure + mysolar->autoExposureStep <= mysolar->autoExposureRange[1])
						{
							exposure += mysolar->autoExposureStep;
						}
						else
						{
							exposure = mysolar->autoExposureRange[1];
						}
						if(exposure != mysolar->attr_ExposureTime_write)
						{
							DEBUG_STREAM << "readthread::run(): AutoExposure: increasing to "<<exposure<<endl;
							//mysolar->m_usbCCD.SetExpTime((float)exposure);
#if _SEABREEZE_API > 1
							unsigned long integr_time_usec = (unsigned long)(exposure*1000);
							sbapi_spectrometer_set_integration_time_micros (mysolar->device_id, mysolar->spectrometer_id, &error, integr_time_usec);
#else
							long integr_time_usec = (long)(exposure*1000);
						    seabreeze_set_integration_time_microsec(mysolar->device_id, &error, integr_time_usec);
#endif
							mysolar->attr_ExposureTime_write = exposure;
						}
					}
				}
				//^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
				//:::::::::::::::::::::: Gaussian Fitting ::::::::::::::::::::::::
				//________________________________________________________________
				if(*(mysolar->attr_GaussFitEnable_read))
				{
					int ret_fun;
					double fit[CONFITEOR_FIT_NUM_PARAM];
					//if last fit ok, initial guess is last fit
					if(mysolar->GaussFit_valid || use_last_fit)
					{
						fit[0] = mysolar->attr_GaussFit_read[0];
						fit[1] = (mysolar->attr_GaussFit_read[1] > 0) ? mysolar->attr_GaussFit_read[1] : 
							((*(mysolar->attr_FWHMEnable_read) && (mysolar->attr_FWHM_read[1] > 0)) ? 
							mysolar->attr_FWHM_read[1] : 50 + (10*counter % 950));
						int center_ind=(int)(((double)(mysolar->AnalysisRegionEndPix - mysolar->AnalysisRegionStartPix + 1))/2);
						//if center is inside analysis region use it, otherwise use fwhm, otherwise use center of analysis region
						fit[2] = (mysolar->attr_GaussFit_read[2] > mysolar->x_values[mysolar->AnalysisRegionStartPix] && mysolar->attr_GaussFit_read[2] < mysolar->x_values[mysolar->AnalysisRegionEndPix]) ? 
							mysolar->attr_GaussFit_read[2] : ((*(mysolar->attr_FWHMEnable_read) && (mysolar->attr_FWHM_read[0] > 0)) ? mysolar->attr_FWHM_read[0] : mysolar->x_values[center_ind]);
						double width = (mysolar->x_values[mysolar->AnalysisRegionEndPix] - mysolar->x_values[mysolar->AnalysisRegionStartPix]);

						double width_fit_uncalibr;
						if(*(mysolar->attr_WidhtCorrectionEnable_read))
						{
							width_fit_uncalibr=sqrt(mysolar->attr_GaussFit_read[3]*mysolar->attr_GaussFit_read[3] - mysolar->widthCorrection*mysolar->widthCorrection);
						}
						else
						{
							width_fit_uncalibr=mysolar->attr_GaussFit_read[3];
						}

						//if rms computed is too big, use last fwhm, otherwise use a fraction of analysis region
						fit[3] = (width_fit_uncalibr > 0 && width_fit_uncalibr < width) ? width_fit_uncalibr*SIGMA_TO_FWHM_FACTOR :
							((*(mysolar->attr_FWHMEnable_read) && (mysolar->attr_FWHM_read[2] > 0)) ? mysolar->attr_FWHM_read[2]*SIGMA_TO_FWHM_FACTOR : width/10.0);
						fit[4] = fit[3];
						DEBUG_STREAM << __func__<<": initial guess with GaussFit_valid"<<endl;
					}
					//otherwise if center and width reference are defined, use that (but not always, maybe they are set to bad values)
					else if ((mysolar->attr_GaussReference_read[2]) != 0 && (mysolar->attr_GaussReference_read[3]) && (counter % 10 == 0))
					{
						fit[0] = mysolar->attr_GaussReference_read[0];
						fit[1] = mysolar->attr_GaussReference_read[1];
						fit[2] = mysolar->attr_GaussReference_read[2];

						double width_ref_uncalibr;
						if(*(mysolar->attr_WidhtCorrectionEnable_read))
						{
							width_ref_uncalibr=sqrt(mysolar->attr_GaussReference_read[3]*mysolar->attr_GaussReference_read[3] - mysolar->widthCorrection*mysolar->widthCorrection);
						}
						else
						{
							width_ref_uncalibr=mysolar->attr_GaussReference_read[3];
						}
						fit[3] = width_ref_uncalibr*SIGMA_TO_FWHM_FACTOR;		//TODO: user is aware of using fwhm instead of sigma?
						fit[4] = fit[3];
						DEBUG_STREAM << __func__<<": initial guess from attr_GaussReference_read"<<endl;
					}
					//otherwise start with some default values
					else
					{
						//fit[0] = mysolar->noise_level; //TODO:
//#ifdef _NON_AVERAGE
//						fit[0] =  //TODO:
//#else
						//average on 4 points: (2 at beginning + 2 at end)/4
						fit[0] = (mysolar->pLiveData_avg[2 + mysolar->AnalysisRegionStartPix] +
							mysolar->pLiveData_avg[5 + mysolar->AnalysisRegionStartPix] +
							mysolar->pLiveData_avg[mysolar->AnalysisRegionEndPix -3] +
							mysolar->pLiveData_avg[mysolar->AnalysisRegionEndPix -6])/4;
//#endif
						//try the range 50...10000 as amplitude
						fit[1] = (mysolar->attr_FWHM_read[1] > 0 && *(mysolar->attr_FWHMEnable_read)) ? mysolar->attr_FWHM_read[1] : 50 + (10*counter % 950);
						//if don't have a value from FWHM, use as center the middle of analysis region
						int analysis_region_size = mysolar->AnalysisRegionEndPix - mysolar->AnalysisRegionStartPix + 1;
						int center_ind=(int)(((double)(analysis_region_size + ((-1 + 2*(counter%2)) * counter%(analysis_region_size/5))))/2);
						fit[2] = (mysolar->attr_FWHM_read[0] > 0 && *(mysolar->attr_FWHMEnable_read)) ? mysolar->attr_FWHM_read[0] : mysolar->x_values[center_ind];
						//if don't have a value from FWHM, use as rms a fraction of analysis region
						double width = (mysolar->x_values[mysolar->AnalysisRegionEndPix] - mysolar->x_values[mysolar->AnalysisRegionStartPix]) / (10.0 + ((1 - 2*(counter%2)) * counter%9));
						fit[3] = (mysolar->attr_FWHM_read[2] > 0 && *(mysolar->attr_FWHMEnable_read)) ? mysolar->attr_FWHM_read[2]*SIGMA_TO_FWHM_FACTOR : width;
						fit[4] = (mysolar->attr_FWHM_read[2] > 0 && *(mysolar->attr_FWHMEnable_read)) ? mysolar->attr_FWHM_read[2]*SIGMA_TO_FWHM_FACTOR : width;
						DEBUG_STREAM << __func__<<": initial guess GENERIC center_ind="<<center_ind<<endl;
					}
					fit[5] = 2;
					fit[6] = 2;
					//DEBUG_STREAM << "readthread::run(): gaussfit initial guess="<<fit[0]<<","<<fit[1]<<","<<fit[2]<<","<<fit[3]<<","<<endl;
					double *data = new double[mysolar->AnalysisRegionEndPix - mysolar->AnalysisRegionStartPix + 1];
					for(unsigned int i=mysolar->AnalysisRegionStartPix; i <= mysolar->AnalysisRegionEndPix; i++)
					{
						data[i-mysolar->AnalysisRegionStartPix] = mysolar->pLiveData_avg[i];
					}
					//size_t fitting_iter;
					unsigned int fitting_iter;
					//SYSTEMTIME ts_end, ts_start;
					//GetSystemTime(&ts_start);
//					ret_fun = (mysolar->m_gaussfit)->perform(data,mysolar->AnalysisRegionEndPix - mysolar->AnalysisRegionStartPix + 1, MAX_GAUSS_ITERATION, fitting_iter, &(mysolar->x_values[mysolar->AnalysisRegionStartPix]), fit);
					double tmperr[CONFITEOR_FIT_NUM_PARAM], tmpchi;
					double mean, rms;
					int ret_fun2;
					timespec start, now;
					clock_gettime(CLOCK_MONOTONIC,&start);
					switch(*(mysolar->attr_FitType_read))
					{
						case GAUSS_FIT:
						default:
							ret_fun = (mysolar->m_gaussfit)->perform(data,&(mysolar->x_values[mysolar->AnalysisRegionStartPix]), mysolar->AnalysisRegionEndPix - mysolar->AnalysisRegionStartPix + 1, MAX_GAUSS_ITERATION, 0, fitting_iter, NULL/*weights*/, fit, tmperr, tmpchi);
							ret_fun2 = (mysolar->m_gaussfit)->compute_rms(fit, &mean, &rms);
							break;
						case GAUSS_FIT_ASYMM:
							ret_fun = (mysolar->m_asyngaussfit)->perform(data,&(mysolar->x_values[mysolar->AnalysisRegionStartPix]), mysolar->AnalysisRegionEndPix - mysolar->AnalysisRegionStartPix + 1, MAX_GAUSS_ITERATION, 0, fitting_iter, NULL/*weights*/, fit, tmperr, tmpchi);
							ret_fun2 = (mysolar->m_asyngaussfit)->compute_rms(fit, &mean, &rms);
							break;
						case CONFITEOR_FIT:
							ret_fun = (mysolar->m_confiteorfit)->perform(data,&(mysolar->x_values[mysolar->AnalysisRegionStartPix]), mysolar->AnalysisRegionEndPix - mysolar->AnalysisRegionStartPix + 1, MAX_CONFITEOR_ITERATION, 0, fitting_iter, NULL/*weights*/, fit, tmperr, tmpchi);
							ret_fun2 = (mysolar->m_confiteorfit)->compute_rms(fit, &mean, &rms);
							break;
					}
					delete [] data;
					clock_gettime(CLOCK_MONOTONIC,&now);
					//printf ("gauss_f: f[0]=%f f[50]=%f\n",f[0],f[50]);
					double fit_time =(now.tv_sec - start.tv_sec)*1000000 + ((double)(now.tv_nsec - start.tv_nsec))/1000;
					fit_counter++;
					if(fit_time < min_fit_time)
						min_fit_time = fit_time;
					if(fit_time > max_fit_time)
						max_fit_time = fit_time;
					tot_fit_time += fit_time;
					//if(fit_counter > 1)
						avg_fit_time = tot_fit_time/(fit_counter);
					//GetSystemTime(&ts_end);
					//DEBUG_STREAM << "readthread::run(): gaussfit performing time: start="<<ts_start.wSecond<<"."<<ts_start.wMilliseconds<<" end="<<ts_end.wSecond<<"."<<ts_end.wMilliseconds<<endl;
					//if(ret_fun)
					bool fit_valid = false;
					//TODO: adjust rms thresh with fwhm
					//eg: if(*(mysolar->attr_FWHMEnable_read))
					//rms_thresh=fabs(mysolar->attr_FWHM_read[2]*SIGMA_TO_FWHM_FACTOR/2)
					if(ret_fun >= 0 && fabs(rms) > 0.08)
					{
						use_last_fit = true;
						mysolar->attr_GaussFit_read[0] = fit[0];
						mysolar->attr_GaussFit_read[1] = fit[1];
						//mysolar->attr_GaussFit_read[2] = fit[2];
						mysolar->attr_GaussFit_read[2] = mean;
						//double temp = fit[3]/(double)(SIGMA_TO_FWHM_FACTOR);
						double temp = rms/(double)(SIGMA_TO_FWHM_FACTOR);
						if(*(mysolar->attr_WidhtCorrectionEnable_read))
						{
							mysolar->attr_GaussFit_read[3]=sqrt(temp*temp + mysolar->widthCorrection*mysolar->widthCorrection);
						}
						else
						{
							mysolar->attr_GaussFit_read[3] = (temp > 0) ? temp : -temp;
						}
					}
					else
					{
						use_last_fit = false;
					}
					if(!ret_fun && !ret_fun2)
					{
						//if constraints amp > 0, mean inside analysis region and rms > 1
						if(fit[1] > 0 && mean > mysolar->x_values[mysolar->AnalysisRegionStartPix] && mean < mysolar->x_values[mysolar->AnalysisRegionEndPix] && fabs(rms) > 0.08)
							fit_valid = true;
					}
					if(fit_valid)
					{
						mysolar->attr_GaussFit_read[0] = fit[0];
						mysolar->attr_GaussFit_read[1] = fit[1];
						//mysolar->attr_GaussFit_read[2] = fit[2];
						mysolar->attr_GaussFit_read[2] = mean;
						//double temp = fit[3]/(double)(SIGMA_TO_FWHM_FACTOR);
						double temp = rms/(double)(SIGMA_TO_FWHM_FACTOR);
						if(*(mysolar->attr_WidhtCorrectionEnable_read))
						{
							mysolar->attr_GaussFit_read[3]=sqrt(temp*temp + mysolar->widthCorrection*mysolar->widthCorrection);
						}
						else
						{
							mysolar->attr_GaussFit_read[3] = (temp > 0) ? temp : -temp;
						}
						*(mysolar->attr_GaussFitCenter_read) = mysolar->attr_GaussFit_read[2];
						*(mysolar->attr_GaussFitWidth_read) = mysolar->attr_GaussFit_read[3];
						double val_buff;
						val_buff = *(mysolar->attr_GaussFitCenter_read);
						mysolar->tbuf[BUF_ID_GAUSS_CENTER].push_data(&val_buff,buf_time,buf_bunch_num);
						val_buff = *(mysolar->attr_GaussFitWidth_read);
						mysolar->tbuf[BUF_ID_GAUSS_WIDTH].push_data(&val_buff,buf_time,buf_bunch_num);
						mysolar->GaussFit_valid = true;
						if(!(*(mysolar->attr_FWHMEnable_read)) || !mysolar->preferFWHM)
							*(mysolar->attr_Center_read) = mysolar->attr_GaussFit_read[2];
							*(mysolar->attr_Width_read) = mysolar->attr_GaussFit_read[3];
						DEBUG_STREAM << "readthread::run(): FIT VALID! fit_time="<<fit_time << "["<<min_fit_time<<","<<max_fit_time<<"]->"<<avg_fit_time << "    fit[1]="<<fit[1]<<" mean="<<mean<<" rms="<<rms;
					}
					else
					{
						DEBUG_STREAM << "readthread::run(): FIT NOT VALID: ret_fun="<<ret_fun<<" ret_fun2="<<ret_fun2<<" fit[1]="<<fit[1]<<" mean="<<mean<<" rms="<<rms<<" fit_time="<<fit_time << "["<<min_fit_time<<","<<max_fit_time<<"]->"<<avg_fit_time;
						mysolar->GaussFit_valid = false;
					}

				}
				//^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
				//:::::::::::::::::::::::  AREA processing :::::::::::::::::::::::
				//________________________________________________________________
				if(*(mysolar->attr_AreaEnable_read))
				{
					size_t x_start, x_end;
					double wl_start, wl_end;
					bool region_ok = false;
					if((*(mysolar->attr_FWHMEnable_read) || (*(mysolar->attr_GaussFitEnable_read) && mysolar->GaussFit_valid)) &&
						*(mysolar->attr_Center_read) > mysolar->attr_RegionOfInterest_read[0] &&
						*(mysolar->attr_Center_read) < mysolar->attr_RegionOfInterest_read[1] && !mysolar->areaInAnalysisRegion)
					{
						wl_start = *(mysolar->attr_Center_read) - (*(mysolar->attr_AreaRegion_read)/2);
						if(wl_start < mysolar->attr_RegionOfInterest_read[0])
							wl_start = mysolar->attr_RegionOfInterest_read[0];
						wl_end = *(mysolar->attr_Center_read) + (*(mysolar->attr_AreaRegion_read)/2);
						if(wl_end > mysolar->attr_RegionOfInterest_read[1])
							wl_end = mysolar->attr_RegionOfInterest_read[1];
						x_start = mysolar->WaveLenght2Pix(wl_start);
						x_end = mysolar->WaveLenght2Pix(wl_end);
						region_ok = true;
					}
					else if(mysolar->attr_AnalysisRegion_read[0] >= mysolar->attr_RegionOfInterest_read[0] &&
						mysolar->attr_AnalysisRegion_read[1] <= mysolar->attr_RegionOfInterest_read[1] && mysolar->areaInAnalysisRegion)
					{
						wl_start = mysolar->attr_AnalysisRegion_read[0];
						wl_end = mysolar->attr_AnalysisRegion_read[1];
						x_start = mysolar->WaveLenght2Pix(wl_start);
						x_end = mysolar->WaveLenght2Pix(wl_end);
						region_ok = true;
					}

					if(region_ok)
					{
						double area=0;
						for(size_t x = x_start; x <= x_end; x++)
							area += fabs(mysolar->pLiveData_avg[x]);
						*(mysolar->attr_Area_read) = area;
						mysolar->tbuf[BUF_ID_AREA].push_data(&area,buf_time,buf_bunch_num);
					}
				}
				//^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
				//:::::::::::::::: End of Single Mode Acquirement ::::::::::::::::
				//________________________________________________________________
				if(mysolar->measure_mode)
				{
					//mutex_sleep->lock();
					pausesec = 1;
					pausenano = 0;		
					//mutex_sleep->unlock();
					mysolar->init_measure_mode = false;
					mysolar->measure_mode = false;
					mysolar->set_state(Tango::ON);
					mysolar->set_status("Data ready");				
				}
				/*ret= mysolar->m_usbCCD.StartMeasuring(&(pNoiseData[mysolar->noise_frames_counter * mysolar->m_usbCCD.m_nNumReadOuts * mysolar->NumPixel]));
				if(!ret)
				{
					INFO_STREAM << "readthread::run(): error in StartMeasuring()" << endl;
				}*/
			}
#if 0
			else if(stat == STATUS_WAIT_DATA)
			{
				if((wait_counter % 50) == 0)
					DEBUG_STREAM << "readthread::run() WAIT_DATA!! wait_counter=" << wait_counter << endl;
				abort_sleep(1.0/100);		//TODO: 10 msec, 
				wait_counter++;
				if(wait_counter > 1000)		//10 sec
				{
					mysolar->live_mode = false;
					mysolar->init_live_mode = false;
					mysolar->init_measure_mode = false;
					mysolar->measure_mode = false;
					pausesec = 1;
					pausenano = 0;
					mysolar->set_state(Tango::ON);
					mysolar->set_status("Timeout waiting data");
					if(!mysolar->m_usbCCD.Reset())
					{
						TangoSys_MemStream out_stream;
						out_stream << "Error resetting CCD" << ends;
						INFO_STREAM << "readthread::run(): "  << out_stream.str() << endl;
					}
				}
			}
			else if(stat == STATUS_WAIT_TRIG)
			{
				if((wait_counter % 50) == 0)
					DEBUG_STREAM << "readthread::run() WAIT_TRIG!! wait_counter=" << wait_counter << endl;
				abort_sleep(1.0/100);		//TODO: 10 msec, 
				wait_counter++;
				if(wait_counter > 1000)		//10 sec
				{
					mysolar->live_mode = false;
					mysolar->init_live_mode = false;
					mysolar->init_measure_mode = false;
					mysolar->measure_mode = false;
					pausesec = 1;
					pausenano = 0;
					mysolar->set_state(Tango::ON);
					mysolar->set_status("Timeout waiting trigger");
					if(!mysolar->m_usbCCD.Reset())
					{
						TangoSys_MemStream out_stream;
						out_stream << "Error resetting CCD" << ends;
						INFO_STREAM << "readthread::run(): "  << out_stream.str() << endl;
					}
				}
			}
#endif
		}
		//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
		//++++++++++++++++++++++++ACQUIRE NOISE MODE+++++++++++++++++++++++++++//
		//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++//
		if(mysolar->acquire_noise_mode)
		{
			DWORD stat;
			//mysolar->m_usbCCD.GetMeasStatus(&stat);
			//if(stat == STATUS_DATA_READY)
			if(1)
			{
				wait_counter = 0;
				//mysolar->m_usbCCD.GetData(&(pNoiseData[mysolar->noise_frames_counter * mysolar->m_usbCCD.m_nNumReadOuts * mysolar->NumPixel]));
#if _SEABREEZE_API > 1
				/*TODO: length*/flag = sbapi_spectrometer_get_formatted_spectrum(mysolar->device_id, mysolar->spectrometer_id, &error, &(pNoiseData[mysolar->noise_frames_counter * mysolar->NumPixel]), (int)mysolar->NumPixel);
#else
				flag = seabreeze_get_formatted_spectrum(mysolar->device_id, &error, &(pNoiseData[mysolar->noise_frames_counter * mysolar->NumPixel]), (int)mysolar->NumPixel);
#endif
				mysolar->noise_frames_counter++;
				DEBUG_STREAM << "readthread::run() DATA_READY!! noise_frames_counter="<<mysolar->noise_frames_counter <<" noise_frames="<<mysolar->noise_frames << endl;
				if(mysolar->noise_frames_counter >= mysolar->noise_frames)
				{
					double noise_avg=0;
					double min_val=0, max_val=10000;
					int tmp_count=0;
					//for every pixel
					for(unsigned int j=0; j < mysolar->noise_frames * mysolar->NumPixel; j++)
					{
						if(pNoiseData[j] >= min_val && pNoiseData[j] <= max_val)
						{
							noise_avg += pNoiseData[j];
							tmp_count++;
						}
					}
					DEBUG_STREAM << "readthread::run() DATA_READY!! noise sum="<<noise_avg <<" noise pixel counter="<<tmp_count<<" noise pixel total=" << mysolar->noise_frames * mysolar->NumPixel<< endl;
					//divide the sum per the number of frames
					noise_avg = noise_avg/((double)tmp_count);
					//noise_avg = noise_avg/((double)mysolar->noise_frames * mysolar->m_usbCCD.m_nNumReadOuts * mysolar->NumPixel);
					mysolar->noise_level = (int)noise_avg; //truncate
					mysolar->attr_NoiseLevel_write = mysolar->noise_level;
					
					for(size_t k=0; k<mysolar->NumPixel; k++)
					{
						noise_avg=0;
						for(size_t j=0; j<mysolar->noise_frames; j++)
						{
							noise_avg += pNoiseData[j* mysolar->NumPixel + k];
						}
						mysolar->pNoiseMask_avg[k] = noise_avg/mysolar->noise_frames;
					}

					ofstream  fd;		//startup file
					fd.open(mysolar->noiseProfileFile.c_str());
					if (!fd)
					{
						stringstream tmp;
						tmp << "Cannot open Noise Profile default file " << mysolar->noiseProfileFile<< "  err:" << strerror(errno) << endl;
						INFO_STREAM << "readthread::acquire_noise(): " << tmp.str();
				   	}
					else
					{
						DEBUG_STREAM << "readthread::acquire_noise(): file " << mysolar->noiseProfileFile << " opened successfully" << endl;
						for(int i=0; i<mysolar->NumPixel; i++)
						{
							fd << mysolar->pNoiseMask_avg[i] << endl;
						}
						fd.close();
					}

					try
					{
						Tango::MultiAttribute *attr_list;
						attr_list = mysolar->get_device_attr();
						Tango::WAttribute &attr_l = attr_list->get_w_attr_by_name ("NoiseLevel");
						attr_l.set_write_value(mysolar->attr_NoiseLevel_write);
					} catch(...)
					{
						DEBUG_STREAM << "readthread::run(): failed setting write value of 'NoiseLevel'";
					}
					DEBUG_STREAM << "readthread::run(): calculated noise average="<<noise_avg<<" -> noise_level="<<mysolar->noise_level<<endl;
					
					delete [] pNoiseData;
					pNoiseData=NULL;

					//mutex_sleep->lock();
					pausesec = 1;
					pausenano = 0;		
					//mutex_sleep->unlock();
					mysolar->init_acquire_noise_mode = false;
					mysolar->acquire_noise_mode = false;
					mysolar->set_state(Tango::ON);
					mysolar->set_status("Data ready");
				}
				/*else
				{
					ret= mysolar->m_usbCCD.StartMeasuring(&(pNoiseData[mysolar->noise_frames_counter * mysolar->m_usbCCD.m_nNumReadOuts * mysolar->NumPixel]));
					if(!ret)
					{
						INFO_STREAM << "readthread::run(): error in StartMeasuring()" << endl;
					}
				}*/
			}
#if 0
			else if(stat == STATUS_WAIT_DATA)
			{
				if((wait_counter % 50) == 0)
					DEBUG_STREAM << "readthread::run() WAIT_DATA!! noise_frames_counter="<<mysolar->noise_frames_counter <<" noise_frames="<<mysolar->noise_frames << " wait_counter="  << wait_counter<< endl;
				abort_sleep(1.0/100);		//TODO: 10 msec, 
				wait_counter++;
				if(wait_counter > 1000)		//10 sec
				{
					mysolar->init_acquire_noise_mode = false;
					mysolar->acquire_noise_mode = false;
					pausesec = 1;
					pausenano = 0;
					mysolar->set_state(Tango::ON);
					mysolar->set_status("Timeout waiting noise data");
					if(!mysolar->m_usbCCD.Reset())
					{
						TangoSys_MemStream out_stream;
						out_stream << "Error resetting CCD" << ends;
						INFO_STREAM << "readthread::run(): "  << out_stream.str() << endl;
					}
				}
			}
			else if(stat == STATUS_WAIT_TRIG)
			{
				if((wait_counter % 50) == 0)
					DEBUG_STREAM << "readthread::run() WAIT_TRIG!! wait_counter=" << wait_counter << endl;
				abort_sleep(1.0/100);		//TODO: 10 msec, 
				wait_counter++;
				if(wait_counter > 1000)		//10 sec
				{
					mysolar->init_acquire_noise_mode = false;
					mysolar->acquire_noise_mode = false;
					pausesec = 1;
					pausenano = 0;
					mysolar->set_state(Tango::ON);
					mysolar->set_status("Timeout waiting trigger");
					if(!mysolar->m_usbCCD.Reset())
					{
						TangoSys_MemStream out_stream;
						out_stream << "Error resetting CCD" << ends;
						INFO_STREAM << "readthread::run(): "  << out_stream.str() << endl;
					}
				}
			}
#endif
		}
		if(((mysolar->measure_mode || mysolar->live_mode || mysolar->acquire_noise_mode) && ((counter % 100) == 0)) ||
			(!mysolar->measure_mode && !mysolar->live_mode && !mysolar->acquire_noise_mode))
		{
			//DEBUG_STREAM << "readthread::run(): measure_mode=" << mysolar->measure_mode << " mysolar->live_mode=" << mysolar->live_mode << endl;
			//DEBUG_STREAM << "readthread::run(): pausesec=" << pausesec << " pausenano=" << pausenano << endl;
			//DEBUG_STREAM << "readthread::run(): before GetParameters()" << endl;
#if 0
			if(!mysolar->m_usbCCD.HitTest())
			{
				INFO_STREAM << "readthread::run(): error in HitTest()" << endl;
				mysolar->set_state(Tango::FAULT);
				mysolar->set_status("Error communicating with CCD");
				pausesec = 1;
				pausenano = 0;
				mysolar->live_mode = false;
				mysolar->init_live_mode = false;
				mysolar->init_measure_mode = false;
				mysolar->measure_mode = false;
				mysolar->init_acquire_noise_mode = false;
				mysolar->acquire_noise_mode = false;
				if(mysolar->m_usbCCD.InitInstance())
				{
					mysolar->m_usbCCD.Attach(mysolar->deviceId);
					mysolar->set_state(Tango::ON);
					mysolar->set_status("Ready");
					mysolar->init_data();
				}
			}
			if(!mysolar->m_usbCCD.GetParameters())
			{
				//INFO_STREAM << "readthread::run(): error in GetParameters()" << endl;		
				if(mysolar->m_usbCCD.InitInstance())
				{
					mysolar->m_usbCCD.Attach(mysolar->deviceId);
					mysolar->set_state(Tango::ON);
					mysolar->set_status("Ready");
					mysolar->init_data();
				}
			}
#endif
		}
		abort_sleep(pausesec + ((double)pausenano)/1000000000);
		counter++;	
	}
	INFO_STREAM << "readthread::run(): exit!!!" << endl;
}

//+------------------------------------------------------------------
//
//	method:			readthread::abort_sleep
//
//	description:	Resume from sleep if abort_flag set (sec.)
//
//-------------------------------------------------------------------
void readthread::abort_sleep(double time) 
{		
	for (int i = 0; i < (time/0.1); i++) {
		if (mysolar->abortflag)
			break;
		if(mysolar->init_live_mode)
			break;
		if(mysolar->init_measure_mode)
			break;
		omni_thread::sleep(0,100000000);
	}
}



